

/*
 *	imgmath.c: Image math functions
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <stdio.h>
#include <string.h>
#include <math.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "application.h"
#include "defs.h"
#include "dlgbox.h"
#include "filter.h"
#include "fft.h"
#include "functions.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "imgmath.h"
#include "interface.h"
#include "roi.h"
#include "secondwnd.h"


/* After any operation that changes the lowest-level stack image (mainimg),
	we need to update the display */

void im_updatedisplay ()
{

	make_minmax (&mainimg);
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);
}



void im_push ()		/* copy main image into first stack register (x) */
{

	if (im_stack.x.data) freebuf (im_stack.x.data);
	if (clone_image (&mainimg, &im_stack.x)==-1)
		fatal_memerr ();

}


void im_swap ()		/* Exchange x-register with main image */
{
image_type tempimg;

	tempimg = mainimg;
	mainimg=im_stack.x;
	im_stack.x=tempimg;

	im_updatedisplay ();

}


/* Note: To prevent range errors, all result images are of type float.
	The exception is RGB, where we rescale the image to fit 8-bit */

/* Note: Size match checking has been done before. All images are supposed 
	to have the same x/y size as well as compatible types */


float im_op (float a, float b, int fn)
{
	if (fn==1) b += a;
	else if (fn==2) b -= a;
	else if (fn==3) b *= a;
	else if (fn==4)
	{
		if (a==0.0) b=0.0; else b /= a;
	}
	else if (fn==5) b = sqrt(ARITHM(a,b));
	else if (fn==6) b = SQR(a-b);
	return b;
}


unsigned short im_bool_op (unsigned short a, unsigned short b, int fn)
{
	if (fn==1) b &= a;
	else if (fn==2) b |= a;
	else if (fn==3) b ^= a;
	return b;
}


int im_rescale (float val, float min, float max)
{
int i;

	val = 255.0*(val-min)/(max-min);
	i = (int) (0.5+val);
	if (i<0) i=0; else if (i>255) i=255;
	return i;
}
	
	

void im_addsubmuldiv (image_type *img1, image_type *img2, int fn)		/* img1 = img1 op img2 */
{
int x,y,z,z1,z2,e;
float *res;
image_type result;
float buf1,buf2;


	if (img1->imgtype==B_RGB)
	{
		image_type *imga, *imgb;
		rgbtriplet *pa, *pb;
		
		if (img1->zmax >= img2->zmax)
		{
			imga=img1; imgb=img2;		/* Make sure that imga is the stack, imgb is the single slice */
		}
		else
		{
			imga=img2; imgb=img1;
		}
		pa = imga->data; pb = imgb->data;

		if (allocate_image (&result, B_FLOAT, 3*img1->xmax, img1->ymax, imga->zmax)==-1)
			fatal_memerr ();
		res = result.data;

		for (y=0; y<imga->ymax; y++)
			for (x=0; x<imga->xmax; x++)
				for (z=0; z<imga->zmax; z++)
				{
					z1=MIN(z,imgb->zmax);
					
					buf1 = pa[x + imga->xmax*(y+imgb->ymax*z) ].red;
					buf2 = pb[x + imga->xmax*(y+imgb->ymax*z1) ].red;
					res[3*x + result.xmax*(y+result.ymax*z) ] = im_op (buf1, buf2, fn);
					
					buf1 = pa[x + imga->xmax*(y+imgb->ymax*z) ].green;
					buf2 = pb[x + imga->xmax*(y+imgb->ymax*z1) ].green;
					res[1+3*x + result.xmax*(y+result.ymax*z) ] = im_op (buf1, buf2, fn);
					
					buf1 = pa[x + imga->xmax*(y+imgb->ymax*z) ].blue;
					buf2 = pb[x + imga->xmax*(y+imgb->ymax*z1) ].blue;
					res[2+3*x + result.xmax*(y+result.ymax*z) ] = im_op (buf1, buf2, fn);
				}
		
		/* Rescale image */
		
		make_minmax (&result);
		imgminmax (&result, &buf1, &buf2);
		for (y=0; y<imga->ymax; y++)
			for (x=0; x<imga->xmax; x++)
				for (z=0; z<imga->zmax; z++)
				{
					e = im_rescale (readbuf_flt (&result,3*x,y,z), buf1, buf2);
					pa[x + imga->xmax*(y+imgb->ymax*z) ].red = e;
					e = im_rescale (readbuf_flt (&result,1+3*x,y,z), buf1, buf2);
					pa[x + imga->xmax*(y+imgb->ymax*z) ].green = e;
					e = im_rescale (readbuf_flt (&result,2+3*x,y,z), buf1, buf2);
					pa[x + imga->xmax*(y+imgb->ymax*z) ].blue = e;
				}
		if (img1!=imga)
			handback (img1, imga);
		else
			img1->uid = next_uid++;
			
		freebuf (&result);
		freebuf (img2);
	
	}
	else
	{
		/* Allocate the result */

		if (allocate_image (&result, B_FLOAT, img1->xmax, img1->ymax, MAX(img1->zmax,img2->zmax))==-1)
			fatal_memerr ();
		res = result.data;

		for (y=0; y<img1->ymax; y++)
			for (x=0; x<img1->xmax; x++)
				for (z=0; z<MAX(img1->zmax,img2->zmax); z++)
				{
					z1=MIN(z,img1->zmax);
					z2=MIN(z,img2->zmax);
					buf1 = readbuf_flt (img1,x,y,z1);
					buf2 = readbuf_flt (img2,x,y,z2);
					res[x + result.xmax*(y+result.ymax*z) ] = im_op (buf1, buf2, fn);
				}
		handback (img1, &result);
		freebuf (img2);			/* kick img2 off the stack */
	}
}


/* Note that boolean ops are not allowed for float. if either image is float, exit */

void im_boolean (image_type *img1, image_type *img2, int fn)		/* img1 = img1 op img2 */
{
int x,y,z,z1,z2,e;
unsigned char *cp;
short *sp;
rgbtriplet *rgbp;
image_type result;
unsigned short buf1,buf2;

	if ((img1->imgtype==B_FLOAT) || (img2->imgtype==B_FLOAT))
		return;

	if ((img1->imgtype==B_RGB) && (img2->imgtype==B_RGB))
	{
		e=3;
		result.imgtype=B_RGB;
	}
	else if ((img1->imgtype==B_SHORT) || (img2->imgtype==B_SHORT))
	{
		e=2;
		result.imgtype=B_SHORT;
	}
	else if ((img1->imgtype==B_CHAR) && (img2->imgtype==B_CHAR))
	{
		e=1;
		result.imgtype=B_CHAR;
	}
	else
		return;


	/* Allocate the result */

	if (allocate_image (&result, result.imgtype, img1->xmax, img1->ymax, MAX(img1->zmax,img2->zmax))==-1)
		fatal_memerr ();
	cp = result.data; sp=result.data; rgbp=result.data;

	for (y=0; y<img1->ymax; y++)
		for (x=0; x<img1->xmax; x++)
			for (z=0; z<MAX(img1->zmax,img2->zmax); z++)
			{
				z1=MIN(z,img1->zmax);
				z2=MIN(z,img2->zmax);
				
				if (e==3)
				{
					buf1 = ((rgbtriplet*)img1->data)[x + img1->xmax*(y+img1->ymax*z) ].red;
					buf2 = ((rgbtriplet*)img2->data)[x + img2->xmax*(y+img2->ymax*z) ].red;
					rgbp[x + result.xmax*(y+result.ymax*z) ].red = im_bool_op (buf1,buf2,fn);
					buf1 = ((rgbtriplet*)img1->data)[x + img1->xmax*(y+img1->ymax*z) ].green;
					buf2 = ((rgbtriplet*)img2->data)[x + img2->xmax*(y+img2->ymax*z) ].green;
					rgbp[x + result.xmax*(y+result.ymax*z) ].green = im_bool_op (buf1,buf2,fn);
					buf1 = ((rgbtriplet*)img1->data)[x + img1->xmax*(y+img1->ymax*z) ].blue;
					buf2 = ((rgbtriplet*)img2->data)[x + img2->xmax*(y+img2->ymax*z) ].blue;
					rgbp[x + result.xmax*(y+result.ymax*z) ].blue = im_bool_op (buf1,buf2,fn);
				}
				else
				{
					if (img1->imgtype==B_CHAR)
						buf1 = ((unsigned char*)img1->data)[x + img1->xmax*(y+img1->ymax*z) ];
					else
						buf1 = ((short*)img1->data)[x + img1->xmax*(y+img1->ymax*z) ];

					if (img2->imgtype==B_CHAR)
						buf2 = ((unsigned char*)img2->data)[x + img2->xmax*(y+img2->ymax*z) ];
					else
						buf2 = ((short*)img2->data)[x + img2->xmax*(y+img2->ymax*z) ];
				
					if (e==2)
						sp[x + result.xmax*(y+result.ymax*z) ] = im_bool_op (buf1,buf2,fn);
					else
						cp[x + result.xmax*(y+result.ymax*z) ] = im_bool_op (buf1,buf2,fn);
				}				

			}
	handback (img1, &result);
	freebuf (img2);			/* kick img2 off the stack */

}



/* General image convolution/deconvolution through FFT 
   -- for convenience we don't accept this for rgb. 
   Note the swapped image1 and image2 parameters. Just for convenience and consistency
*/

void im_genconvolve (image_type *img2, image_type *img1, int fn)		/* img1 = img1 op img2 */
{
int x,y,z,x1,y1;
float *res, *img1p, *img2p;
image_type result, img1clone, img2clone;
float buf1,buf2,buf3, ibuf1, ibuf2, ibuf3;


	if (img1->imgtype==B_RGB)
	{
		return;
	}
	else
	{
		/* Allocate the result */

		if (allocate_image (&result, B_FLOAT, img1->xmax, img1->ymax, MAX(img1->zmax,img2->zmax))==-1)
			fatal_memerr ();
		res = result.data;

		for (z=0; z<MAX(img1->zmax,img2->zmax); z++)
		{
			/* Clone one slice of the imput images */

			if (allocate_image (&img1clone, img1->imgtype, img1->xmax, img1->ymax, 1)==-1)
				fatal_memerr ();
			copybuf_z (img1, &img1clone, z,z);
			if (allocate_image (&img2clone, img1->imgtype, img2->xmax, img2->ymax, 1)==-1)
				fatal_memerr ();
			copybuf_z (img2, &img2clone, z,z);

			/* Transform the input images - transformation funcs in fft.c.
			   The image clones are now 2 slices of floats. */

			dp (1, "Performing forward FFT\n");
		
			fft2d (&img1clone,z);
			fft2d (&img2clone,z);
			img1p = img1clone.data;
			img2p = img2clone.data;
	
			/*  Perform the actual operation:
			
				IMG1*IMG2 convolves img1 with img2
				IMG1/IMG2 deconvolves img1 with img2
				IMG1*conj(IMG2) computes the cross-correlation
			*/

			dp (1, "Performing FFT math, function %d slice %d\n", fn,z);
			for (y=0; y<img1->ymax; y++)
				for (x=0; x<img1->xmax; x++)
				{
					buf1  = img1p[x + img1clone.xmax*y];
					ibuf1 = img1p[x + img1clone.xmax*y + img1clone.xmax*img1clone.ymax];
					buf2  = img2p[x + img2clone.xmax*y];
					ibuf2 = img2p[x + img2clone.xmax*y + img2clone.xmax*img2clone.ymax];
					
					switch (fn)
					{
						case 0:						/* Multiply */
							buf3 = buf1*buf2 - ibuf1*ibuf2;
							ibuf3= buf1*ibuf2 + buf2*ibuf1;
							break;
						case 1:						/* Divide, Bronstein p. 558 */
							buf3 = buf1*buf2 + ibuf1*ibuf2;
							ibuf3= buf2*ibuf1 - buf1*ibuf2;
							buf3 /= SQR(buf2)+SQR(ibuf2);
							ibuf3 /= SQR(buf2)+SQR(ibuf2);
							break;
						case 2:						/* Multiply with conj cplx, Russ pg. 365f */
							buf3 = buf1*buf2 + ibuf1*ibuf2;
							ibuf3= buf2*ibuf1 - buf1*ibuf2;
							break;
						default:
							buf3=ibuf3=0;
					}
					img1p[x + img1clone.xmax*y] = buf3;
					img1p[x + img1clone.xmax*y + img1clone.xmax*img1clone.ymax] = ibuf3;
					img2p[x + img2clone.xmax*y] = buf3;
					img2p[x + img2clone.xmax*y + img2clone.xmax*img2clone.ymax] = ibuf3;
										
				}

			/* Inverse-transform the result and store in the appropriate slice of result
			   The one thing I never understand is why the quadrants are swapped.
			   Anyway, let's swap them back right here. */

			dp (1, "Performing inverse FFT\n");
			ifft2d (&img1clone);
			img1p = img1clone.data;
			for (y=0; y<img1->ymax; y++)
				for (x=0; x<img1->xmax; x++)
				{
					if (x<img1clone.xmax/2) x1=x+img1clone.xmax/2; else x1=x-img1clone.xmax/2;
					if (y<img1clone.ymax/2) y1=y+img1clone.ymax/2; else y1=y-img1clone.ymax/2;
					res[x + y*img1->xmax + z*img1->xmax*img1->ymax] = img1p[x1 + y1*img1clone.xmax];
				}
			dp (1, "Done for slice %d.\n",z);

			/* Free the temporary one-slice clones */

			freebuf (&img1clone);
			freebuf (&img2clone); 

		}
		/*handback (img2, &result);*/
		handback (img2, &result);
		freebuf (img1);			/* kick img2 off the stack */
		dp (1, "Done.\n");
	}
}


/* Stitching / Tiling

	Two images are put side-by-side, either vertically (dir=1) or horizontally (dir=0).
	Strictly, only one dimension must match (x size for vertical stitching or
	y size for horizontal stitching, but we require both to match in order to not
	break the logic of the other operations. The Z size should match, though.
	We are tolerant, however, and use the smaller Z size */


void im_stitch (image_type *img1, image_type *img2, int fn)		/* img1 = img1 op img2 */
{
int x,y,z;
image_type result;

	/* We circumvented size and type checks for this function, so do it here */

	if (img1->imgtype!=img2->imgtype)		/* Image types must match */
	{
		ShowMessage ("Error: Image type mismatch");
		return;
	}
	
	if ((fn==0) && (img1->ymax != img2->ymax))		/* Horizontal tiling w/ Y mismatch */
	{
		ShowMessage ("Error: Y size mismatch");
		return;
	}
	if ((fn==1) && (img1->xmax != img2->xmax))		/* Vertical tiling w/ X mismatch */
	{
		ShowMessage ("Error: X size mismatch");
		return;
	}
	
	if (fn==0)								/* Horizontal stitching */
	{
		if (allocate_image (&result, img1->imgtype, img1->xmax+img2->xmax, img1->ymax, MIN(img1->zmax,img2->zmax))==-1)
			fatal_memerr ();
		
		for (z=0; z<MIN(img1->zmax,img2->zmax); z++)
		{
			switch (img1->imgtype)
			{
				case B_CHAR:
				{
					char *pdest = result.data;
					char *p1 = img1->data;
					char *p2 = img2->data;
					for (y=0; y<img1->ymax; y++)
					{
						for (x=0; x<img1->xmax; x++)
							pdest[x + result.xmax*(y+result.ymax*z)] = p1[x + img1->xmax*(y+img1->ymax*z)];
						for (x=0; x<img2->xmax; x++)
							pdest[img1->xmax+x + result.xmax*(y+result.ymax*z)] = p2[x + img2->xmax*(y+img2->ymax*z)];
					}
					break;
				}
				case B_SHORT:
				{
					short *pdest = result.data;
					short *p1 = img1->data;
					short *p2 = img2->data;
					for (y=0; y<img1->ymax; y++)
					{
						for (x=0; x<img1->xmax; x++)
							pdest[x + result.xmax*(y+result.ymax*z)] = p1[x + img1->xmax*(y+img1->ymax*z)];
						for (x=0; x<img2->xmax; x++)
							pdest[img1->xmax+x + result.xmax*(y+result.ymax*z)] = p2[x + img2->xmax*(y+img2->ymax*z)];
					}
					break;
				}
				case B_FLOAT:
				{
					float *pdest = result.data;
					float *p1 = img1->data;
					float *p2 = img2->data;
					for (y=0; y<img1->ymax; y++)
					{
						for (x=0; x<img1->xmax; x++)
							pdest[x + result.xmax*(y+result.ymax*z)] = p1[x + img1->xmax*(y+img1->ymax*z)];
						for (x=0; x<img2->xmax; x++)
							pdest[img1->xmax+x + result.xmax*(y+result.ymax*z)] = p2[x + img2->xmax*(y+img2->ymax*z)];
					}
					break;
				}
				case B_RGB:
				{
					rgbtriplet *pdest = result.data;
					rgbtriplet *p1 = img1->data;
					rgbtriplet *p2 = img2->data;
					for (y=0; y<img1->ymax; y++)
					{
						for (x=0; x<img1->xmax; x++)
							pdest[x + result.xmax*(y+result.ymax*z)] = p1[x + img1->xmax*(y+img1->ymax*z)];
						for (x=0; x<img2->xmax; x++)
							pdest[img1->xmax+x + result.xmax*(y+result.ymax*z)] = p2[x + img2->xmax*(y+img2->ymax*z)];
					}
					break;
				}
			}
		}
	}
	else		/* vertical stitching */
	{
		if (allocate_image (&result, img1->imgtype, img1->xmax, img1->ymax+img2->ymax, MIN(img1->zmax,img2->zmax))==-1)
			fatal_memerr ();
		
		for (z=0; z<MAX(img1->zmax,img2->zmax); z++)
		{
			switch (img1->imgtype)
			{
				case B_CHAR:
				{
					char *pdest = result.data;
					char *p1 = img1->data;
					char *p2 = img2->data;
					for (x=0; x<img1->xmax; x++)
					{
						for (y=0; y<img1->ymax; y++)
							pdest[x + result.xmax*(y+result.ymax*z)] = p1[x + img1->xmax*(y+img1->ymax*z)];
						for (y=0; y<img2->ymax; y++)
							pdest[x + result.xmax*(y+img1->ymax+result.ymax*z)] = p2[x + img2->xmax*(y+img2->ymax*z)];
					}
					break;
				}
				case B_SHORT:
				{
					short *pdest = result.data;
					short *p1 = img1->data;
					short *p2 = img2->data;
					for (x=0; x<img1->xmax; x++)
					{
						for (y=0; y<img1->ymax; y++)
							pdest[x + result.xmax*(y+result.ymax*z)] = p1[x + img1->xmax*(y+img1->ymax*z)];
						for (y=0; y<img2->ymax; y++)
							pdest[x + result.xmax*(y+img1->ymax+result.ymax*z)] = p2[x + img2->xmax*(y+img2->ymax*z)];
					}
					break;
				}
				case B_FLOAT:
				{
					float *pdest = result.data;
					float *p1 = img1->data;
					float *p2 = img2->data;
					for (x=0; x<img1->xmax; x++)
					{
						for (y=0; y<img1->ymax; y++)
							pdest[x + result.xmax*(y+result.ymax*z)] = p1[x + img1->xmax*(y+img1->ymax*z)];
						for (y=0; y<img2->ymax; y++)
							pdest[x + result.xmax*(y+img1->ymax+result.ymax*z)] = p2[x + img2->xmax*(y+img2->ymax*z)];
					}
					break;
				}
				case B_RGB:
				{
					rgbtriplet *pdest = result.data;
					rgbtriplet *p1 = img1->data;
					rgbtriplet *p2 = img2->data;
					for (x=0; x<img1->xmax; x++)
					{
						for (y=0; y<img1->ymax; y++)
							pdest[x + result.xmax*(y+result.ymax*z)] = p1[x + img1->xmax*(y+img1->ymax*z)];
						for (y=0; y<img2->ymax; y++)
							pdest[x + result.xmax*(y+img1->ymax+result.ymax*z)] = p2[x + img2->xmax*(y+img2->ymax*z)];
					}
					break;
				}
			}
		}
	}

	handback (img1, &result);
	freebuf (img2);			/* kick img2 off the stack */

}


/************************************************************************************************************/

/* Single-op monadic operations (1:abs, 2:log, 3:exp, 4:SQR) */

void monadic_singleop (image_type *img, int fn)
{
int x,y,z;
long l;
image_type result; float *q;
float buf;

	if (fn==0) return;

	if (((img->imgtype==B_CHAR) || (img->imgtype==B_RGB)) && (fn==1))		/* Abs doesn't make sense on uchar image */
	{
		ShowMessage ("Function not valid on unsigned data types");
		return;
	}
	if (img->imgtype==B_RGB)		/* Abs doesn't make sense. Log, exp not good on RGB */
	{
		ShowMessage ("Function not possible on RGB images ");
		return;
	}

	if (fn==1)				/* Handle ABS separately for simplicity. Only B_FLOAT and B_SHORT left */
	{
		if (img->imgtype==B_SHORT)
		{
			short *p = img->data;
			for (l=0; l<img->xmax * img->ymax * img->zmax; l++)
				p[l] = abs(p[l]);
		}
		else if (img->imgtype==B_FLOAT)
		{
			float *p = img->data;
			for (l=0; l<img->xmax * img->ymax * img->zmax; l++)
				p[l] = fabs(p[l]);
		}
		return;
	}

	/* The result of log & exp & SQR will be B_FLOAT */
	
	if (allocate_image (&result, B_FLOAT, img->xmax, img->ymax, img->zmax)==-1)
		fatal_memerr();
	q = result.data;
	
	for (z=0; z<img->zmax; z++)
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				buf = readbuf_flt (img,x,y,z);
				if (fn==2) 			buf = SAFELOG10(buf);
				else if (fn==3) 	buf = pow(10,buf);
				else if (fn==4)		buf = SQR(buf);
				else if (fn==5)		buf = windowfn (buf, (float)(x-img->xmax/2)/(img->xmax/2+2), (float)(y-img->ymax/2)/(img->ymax/2+2),3);
				q[x + img->xmax*(y + img->ymax*z)] = buf;
			}
	
	handback (img, &result);

}


/* Monadic addition, subtraction, multiplication, division with constant operand */

void im_monad_addsubmuldiv (image_type *img, float operand, int fn)
{
int x,y,z,e;
float *res;
rgbtriplet * prgb; rgbtriplet rgb;
image_type result;
float buf, buf1, buf2;


	if (img->imgtype==B_RGB)
	{
		prgb = img->data;

		if (allocate_image (&result, B_FLOAT, 3*img->xmax, img->ymax, img->zmax)==-1)
			fatal_memerr ();
		res = result.data;

		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
				for (z=0; z<img->zmax; z++)
				{
					rgb = readbuf_rgb (img, x,y,z);
					
					buf = rgb.red;
					res[3*x + result.xmax*(y+result.ymax*z) ] = im_op (operand, buf, fn);
					
					buf = rgb.green;
					res[1+3*x + result.xmax*(y+result.ymax*z) ] = im_op (operand, buf, fn);
					
					buf = rgb.blue;
					res[2+3*x + result.xmax*(y+result.ymax*z) ] = im_op (operand, buf, fn);
				}
		
		/* Rescale image */
		
		make_minmax (&result);
		imgminmax (&result, &buf1, &buf2);
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
				for (z=0; z<img->zmax; z++)
				{
					e = im_rescale (readbuf_flt (&result,3*x,y,z), buf1, buf2);
					prgb[x + img->xmax*(y+img->ymax*z) ].red = e;
					e = im_rescale (readbuf_flt (&result,1+3*x,y,z), buf1, buf2);
					prgb[x + img->xmax*(y+img->ymax*z) ].green = e;
					e = im_rescale (readbuf_flt (&result,2+3*x,y,z), buf1, buf2);
					prgb[x + img->xmax*(y+img->ymax*z) ].blue = e;
				}
		img->uid = next_uid++;
		freebuf (&result);
	
	}
	else		/* non-RGB always results in a B_FLOAT image */
	{
		/* Allocate the result */

		if (allocate_image (&result, B_FLOAT, img->xmax, img->ymax, img->zmax)==-1)
			fatal_memerr ();
		res = result.data;

		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
				for (z=0; z<img->zmax; z++)
				{
					buf = readbuf_flt (img,x,y,z);
					res[x + result.xmax*(y+result.ymax*z) ] = im_op (operand, buf, fn);
				}
		handback (img, &result);
	
	}
}



/******************************************/

/*          ***    *   *    ***           */
/*         *   *   *   *     *            */
/*         *       *   *     *            */
/*         *  **   *   *     *            */
/*         *   *   *   *     *            */
/*         *   *   *   *     *            */
/*          ***     ***     ***           */
/*                                        */


GtkWidget* monad_entry;			/* Text entry for constant operand */
int	imgmath_monadfn;



void on_monadfn_cancel_clicked  (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_monadfn_ok_clicked  (GtkButton *button, gpointer user_data)
{
char buf[100];
float oper;

	if (imgmath_monadfn>0) undo_nexus (&mainimg);
	if (imgmath_monadfn>4)							/* Examine the entry field */
	{
		strncpy (buf, gtk_entry_get_text (GTK_ENTRY(monad_entry)),99);
		if ((strlen(buf)==0) || (strchr ("0123456789.-+", buf[0])==NULL))
		{
			ShowMessage ("Illegal Operand");
			return;
		}
		oper = atof (buf);
		if ((oper==0) && (imgmath_monadfn==8))
		{
			ShowMessage ("Division by zero");
			return;
		}
		
	}
	

	switch (imgmath_monadfn)
	{
		case 1:				/* Abs, Log, Exp, and SQR */
		case 2:
		case 3:
		case 4:
				monadic_singleop  (&mainimg, imgmath_monadfn);
				break;
		case 5:				/* Add, Sub, Mul, Div */
		case 6:
		case 7:
		case 8:
				im_monad_addsubmuldiv (&mainimg, oper, imgmath_monadfn-4);
				break;
		case 9:
				monadic_singleop  (&mainimg, 5);
				break;
	}

	im_updatedisplay ();

}


void on_monadic_set_fnmenu (GtkWidget *widget, gpointer user_data)
{

	imgmath_monadfn = *(int*) user_data;
	
}



void im_monadic (image_type *img)
{
GtkWidget* monadicwnd;
GtkWidget* vbox;
GtkWidget *label1, *label2, *label3, *label4;
GtkWidget *table1;
GtkWidget *hsep1, *hsep2;
GtkWidget *im_okbutton, *im_cancelbutton;
GtkWidget *im_fnmenu, *im_fnmenu_menu, *menuitem;



	monadicwnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (monadicwnd), "monadicwnd", monadicwnd);
	gtk_window_set_title (GTK_WINDOW (monadicwnd), "Monadic Functions");
	vbox = vbox_in_window (monadicwnd, "vboxim2");
	table1 = MakeTable (monadicwnd,vbox,"table1a",2,4);
	
	/* Operation menu */
	
	label1 = MakeLabel (monadicwnd,table1,"Operation","label1a",0,0);
	im_fnmenu = Make_Optionmenu (monadicwnd,table1,"im_fnmenu_mon",1,0);
	im_fnmenu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_monadic_set_fnmenu), "Abs (m)", (gpointer) &menuopt1);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_monadic_set_fnmenu), "Log (m)", (gpointer) &menuopt2);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_monadic_set_fnmenu), "Exp (m)", (gpointer) &menuopt3);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_monadic_set_fnmenu), "SQR (m)", (gpointer) &menuopt4);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_monadic_set_fnmenu), "Add (m)+C", (gpointer) &menuopt5);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_monadic_set_fnmenu), "Sub (m)-C", (gpointer) &menuopt6);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_monadic_set_fnmenu), "Mul (m)*C", (gpointer) &menuopt7);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_monadic_set_fnmenu), "Div (m)/C", (gpointer) &menuopt8);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_monadic_set_fnmenu), "Window", (gpointer) &menuopt9);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (im_fnmenu), im_fnmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (im_fnmenu), 0);
	imgmath_monadfn = 1;

	/* Constant operator */

	label2 = MakeLabel (monadicwnd,table1,"Constant Operand","label2a",0,1);
	monad_entry = Entry_in_table (monadicwnd,table1,"","prefs_entry4",1,1);

	hsep1 = MakeHsep (monadicwnd,table1,"hsep1",0,2);
	hsep2 = MakeHsep (monadicwnd,table1,"hsep2",1,2);

	im_okbutton = Button_in_table (monadicwnd,table1,"Execute","mimokbutt",0,3,NULL);
	im_cancelbutton = Button_in_table (monadicwnd,table1,"Dismiss","mimcancelbutt",1,3,NULL);

	/* Callback signals */
	
	gtk_signal_connect (GTK_OBJECT (monadicwnd), "destroy",
                      GTK_SIGNAL_FUNC (on_monadfn_cancel_clicked), monadicwnd);
	gtk_signal_connect (GTK_OBJECT (im_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_monadfn_cancel_clicked), monadicwnd);
	gtk_signal_connect (GTK_OBJECT (im_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_monadfn_ok_clicked), monadicwnd);

	
	gtk_window_set_modal (GTK_WINDOW (monadicwnd), TRUE);	/* Important. Otherwise we could open up this window multiple times */
	gtk_widget_show (monadicwnd);

}


/***************************************************************************************/


int imgmath_function;
GtkWidget *im_label_ximg, *im_label_yimg, *im_label_opinfo;


/* Set the informational labels in the dialog box. Also, return our
	"abilities", i.e. whether we can operate on one image, or we can
	perform dyadic opertors. */

int imgmath_set_info ()
{
char s[256];
int optcode;

	if (mainimg.data)
	{
		sprintf (s, "%lix%lix%li ",mainimg.xmax, mainimg.ymax, mainimg.zmax);
		if (mainimg.imgtype==B_CHAR) strcat (s, "8-bit");
		else if (mainimg.imgtype==B_FLOAT) strcat (s, "IEEE float");
		else if (mainimg.imgtype==B_SHORT) strcat (s, "16-bit");
		else if (mainimg.imgtype==B_RGB) strcat (s, "24-bit RGB");
		optcode=1;
	}
	else
	{
		sprintf (s, "No image");
		optcode=0;
	}

	gtk_label_set_text (GTK_LABEL(im_label_ximg), s);

	if (im_stack.x.data)
	{
		sprintf (s, "%lix%lix%li ", im_stack.x.xmax, im_stack.x.ymax, im_stack.x.zmax);
		if (im_stack.x.imgtype==B_CHAR) strcat (s, "8-bit");
		else if (im_stack.x.imgtype==B_FLOAT) strcat (s, "IEEE float");
		else if (im_stack.x.imgtype==B_SHORT) strcat (s, "16-bit");
		else if (im_stack.x.imgtype==B_RGB) strcat (s, "24-bit RGB");
		if (optcode) optcode++;
	}
	else
	{
		sprintf (s, "No image");
	}
	
	gtk_label_set_text (GTK_LABEL(im_label_yimg), s);
	
	if (optcode==0)
		sprintf (s, "No operation possible");
	else if (optcode==1)
		sprintf (s, "Monadic operations only");
	else
	{			/* Dyadic - check image formats and compatibility  */
	
		if ((mainimg.xmax!=im_stack.x.xmax) || (mainimg.ymax!=im_stack.x.ymax))
		{
			sprintf (s, "Incompatible image size");
			gtk_label_set_text (GTK_LABEL(im_label_opinfo), s);
			if (optcode==2) optcode=1;
			return optcode;
		}			
	
	
		if ((mainimg.zmax==1) && (im_stack.x.zmax==1))
			sprintf (s, "Single-slice operation");
		else if ((mainimg.zmax==1) || (im_stack.x.zmax==1))
			sprintf (s, "Slice on stack");
		else if (mainimg.zmax==im_stack.x.zmax)
			sprintf (s, "Pixel-by-pixel, entire stack");
		else
		{		/* Z size mismach */
			sprintf (s, "Z-size mismatch");
			gtk_label_set_text (GTK_LABEL(im_label_opinfo), s);
			if (optcode==2) optcode=1;
			return optcode;
		}
	
		/* Last check: Both or none are RGB */
		
		if ( ((mainimg.imgtype==B_RGB) && (im_stack.x.imgtype==B_RGB))
				|| ((mainimg.imgtype!=B_RGB) && (im_stack.x.imgtype!=B_RGB)))
		{
			/* Nothing to do - all is fine */
		}
		else
		{
			strcat (s, "\nImage type mismach");
			if (optcode==2) optcode=1;
		}
	
	
	}	
	gtk_label_set_text (GTK_LABEL(im_label_opinfo), s);

	return optcode;
}


void on_imgmath_cancel_clicked  (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


/* Note: To facilitate sequential ops, we do not destroy the window on OK*/

void on_imgmath_ok_clicked  (GtkButton *button, gpointer user_data)
{
int optcode;

	if (imgmath_function>2) undo_nexus (&mainimg); /* NOP and push do not affect mainimage. Monadic has its own undo. */

	optcode = imgmath_set_info ();

	switch (imgmath_function)
	{
		case 1:		if (optcode>0) 
					{
						im_monadic (&mainimg);
						im_updatedisplay ();
					}
					break;
		case 2:		if (optcode>0) im_push ();
					break;
		case 3:		if (optcode>1) im_swap ();
					break;
		case 4:											/* Add/Sub/Mul/Div */
		case 5:
		case 6:
		case 7:
					if (optcode>1)
					{
						im_addsubmuldiv (&mainimg, &im_stack.x, imgmath_function-3);
						im_updatedisplay ();
					}
					break;
		case 8:											/* Boolean - And / Or / Xor */
		case 9:
		case 10:
					if (optcode>1)
					{
						im_boolean (&mainimg, &im_stack.x, imgmath_function-7);
						im_updatedisplay ();
					}
					break;
		case 11:										/* FFT-based convolution, deconvolution, crosscorrelation */
		case 12:
		case 13:
					if (optcode>1)
					{
						im_genconvolve (&mainimg, &im_stack.x, imgmath_function-11);
						im_updatedisplay ();
					}
					break;
					
		case 14:										/* Pixel - by- pixel sqrt (a^2+b^2) */
					if (optcode>1)
					{
						im_addsubmuldiv (&mainimg, &im_stack.x, 5);
						im_updatedisplay ();
					}
					break;
		case 15:
					if ((optcode>0)	/* Count this as monadic and do size checks in the actual function */
						&& (mainimg.data) && (im_stack.x.data))
					{
						im_stitch (&mainimg, &im_stack.x, 0);
						im_updatedisplay ();
					}
					break;
	}


	imgmath_set_info ();
}




void on_imgmath_set_fnmenu (GtkWidget *widget, gpointer user_data)
{

	imgmath_set_info ();
	imgmath_function = *(int*) user_data;
	
}


void create_imgmath_wnd (image_type* img)
{
GtkWidget* imgmathwnd;
GtkWidget* vbox;
GtkWidget *label1, *label2, *label3, *label4;
GtkWidget *table1;
GtkWidget *hsep1, *hsep2;
GtkWidget *im_okbutton, *im_cancelbutton;
GtkWidget *im_fnmenu, *im_fnmenu_menu, *menuitem;



	imgmathwnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (imgmathwnd), "imgmathwnd", imgmathwnd);
	gtk_window_set_title (GTK_WINDOW (imgmathwnd), "Image Math");
	vbox = vbox_in_window (imgmathwnd, "vboxim");
	table1 = MakeTable (imgmathwnd,vbox,"table1",2,6);
	
	/* Three labels indicating the stack state */
	
	label1 = MakeLabel (imgmathwnd,table1,"Image 1 (m)","label1",0,0);
	label2 = MakeLabel (imgmathwnd,table1,"Image 2 (x)","label2",0,1);
	label3 = MakeLabel (imgmathwnd,table1,"Operation","label3",0,2);

	im_label_ximg = MakeLabel (imgmathwnd,table1,"--","label1a",1,0);
	im_label_yimg = MakeLabel (imgmathwnd,table1,"--","label2a",1,1);
	im_label_opinfo = MakeLabel (imgmathwnd,table1,"void","label3a",1,2);

	hsep1 = MakeHsep (imgmathwnd,table1,"hsep1",0,3);
	hsep2 = MakeHsep (imgmathwnd,table1,"hsep2",1,3);
	label4 = MakeLabel (imgmathwnd,table1,"Choose Operation","label4",0,4);

	im_fnmenu = Make_Optionmenu (imgmathwnd,table1,"im_fnmenu",1,4);
	im_fnmenu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "No operation", (gpointer) &menuopt0);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Monadic Operations", (gpointer) &menuopt1);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Push m -> x", (gpointer) &menuopt2);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Swap m <-> x", (gpointer) &menuopt3);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Add x+m -> m", (gpointer) &menuopt4);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Sub x-m -> m", (gpointer) &menuopt5);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Mul x*m -> m", (gpointer) &menuopt6);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Div x/m -> m", (gpointer) &menuopt7);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "And x&m -> m", (gpointer) &menuopt8);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Or  x|m -> m", (gpointer) &menuopt9);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Xor x^m -> m", (gpointer) &menuopt10);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Conv x,m -> m", (gpointer) &menuopt11);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Deconv x,m -> m", (gpointer) &menuopt12);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Crosscorr x,m -> m", (gpointer) &menuopt13);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "Arithm x,m -> m", (gpointer) &menuopt14);
	menuitem = Option_menuitem (im_fnmenu_menu, GTK_SIGNAL_FUNC(on_imgmath_set_fnmenu), "hTile [mx] -> m", (gpointer) &menuopt15);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (im_fnmenu), im_fnmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (im_fnmenu), 0);

	im_okbutton = Button_in_table (imgmathwnd,table1,"Execute","imokbutt",0,5,NULL);
	im_cancelbutton = Button_in_table (imgmathwnd,table1,"Dismiss","imcancelbutt",1,5,NULL);

	/* Callback signals */
	
	gtk_signal_connect (GTK_OBJECT (imgmathwnd), "destroy",
                      GTK_SIGNAL_FUNC (on_imgmath_cancel_clicked), imgmathwnd);
	gtk_signal_connect (GTK_OBJECT (im_cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_imgmath_cancel_clicked), imgmathwnd);
	gtk_signal_connect (GTK_OBJECT (im_okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_imgmath_ok_clicked), imgmathwnd);


	gtk_widget_show (imgmathwnd);
	imgmath_set_info ();

}


void imgmath_main ()
{

	create_imgmath_wnd (&mainimg);
	
}




/**********************************************************/

